/*
 *  Copyright (C) 2017 The Contributors
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *  General Public License for more details.
 *
 *  A copy of the GNU General Public License can be found in the file
 *  "gpl.txt" in the top directory of this repository.
 */

#include <all.h>

#define ver(...) _ver(lcd_verbose, 0, __VA_ARGS__)
#define ver2(...) _ver(lcd_verbose, 1, __VA_ARGS__)
#define ver3(...) _ver(lcd_verbose, 2, __VA_ARGS__)

#define WIN_W (510 * 2)
#define WIN_H (64 * 2)

#define CON_BGR 0x00000000
#define CON_DRW 0xFFFFFFFF
#define CON_FGR ((SDL_Color){ .r = 255, .b = 255, .g = 255, .a = 255 })

#define	G_INIT		0x40
#define	G_MWRITE	0x42
#define	G_MREAD		0x43
#define	G_SETSAD	0x44
#define	G_CRSWR		0x46
#define	G_CRSRD		0x47
#define	G_CRSMRT	0x4C
#define	G_CRSMLT	0x4D
#define	G_CRSMUP	0x4E
#define	G_CRSMDN	0x4F
#define	G_ERASE		0x52
#define	G_SLEEP		0x53
#define	G_DSPOFF	0x58
#define	G_DSPON		0x59
#define	G_HSCRL		0x5A
#define	G_OVRLAY	0x5B
#define	G_CGRAM		0x5C
#define	G_CRSFRM	0x5D

#define TXT_W 85
#define TXT_H 8

#define GFX_W 85
#define GFX_H 64

#define BASE_TXT 0x0000
#define BASE_GFX 0x2000

static uint8_t mem_txt[TXT_H * TXT_W];
static uint8_t mem_gfx[GFX_H * GFX_W];

static SDL_Window *win;
static SDL_Renderer *ren;
static SDL_atomic_t frame;
static SDL_atomic_t clear;

static uint32_t render = 0;

static TTF_Font *fon;
static int32_t fon_w, fon_h;

static int32_t sur_w, sur_h;
static SDL_Surface *sur;
static SDL_Surface *gfx_sur;

static int32_t cur = 0;
static int32_t cur_c = 0;
static int32_t dir = 1;

static int32_t current_op = 0x00;

static uint32_t last_val = 0x00;
static uint32_t last_off = 0;
static int32_t last_sz = -1;
static int32_t mult_val_c = 0;

int32_t lcd_verbose = 0;

void lcd_sdl(void)
{
	ver3("lcd_sdl()");

	if (SDL_AtomicGet(&clear)) {
		if (SDL_SetRenderDrawColor(ren, 0x00, 0x00, 0x00, 0xff) < 0) {
			fail("SDL_SetRenderDrawColor() failed: %s", SDL_GetError());
		}

		if (SDL_RenderClear(ren) < 0) {
			fail("SDL_RenderClear() failed: %s", SDL_GetError());
		}

		SDL_RenderPresent(ren);
		SDL_AtomicSet(&clear, 0);
		return;
	}

	static int32_t last = 0;
	int32_t now = SDL_AtomicGet(&frame);

	if (last == now) {
		ver3("no update");
		return;
	}

	last = now;

	if (SDL_FillRect(sur, NULL, CON_BGR) < 0) {
		fail("SDL_FillRect() failed: %s", SDL_GetError());
	}

	for (int32_t y = 0; y < TXT_H; ++y) {
		char line[TXT_W + 1];

		if (SDL_LockMutex(cpu_mutex) < 0) {
			fail("SDL_LockMutex() failed: %s", SDL_GetError());
		}

		memcpy(line, mem_txt + y * TXT_W, TXT_W);
		line[TXT_W] = 0;

		if (SDL_UnlockMutex(cpu_mutex) < 0) {
			fail("SDL_UnlockMutex() failed: %s", SDL_GetError());
		}

		SDL_Surface *lin = TTF_RenderText_Blended(fon, line, CON_FGR);

		if (lin == NULL) {
			fail("TTF_RenderText_Blended() failed: %s", TTF_GetError());
		}

		if (SDL_BlitSurface(lin, NULL, sur, &(SDL_Rect){
			.x = 0,
			.y = y * fon_h,
			.w = TXT_W * fon_w,
			.h = fon_h
		})) {
			fail("SDL_BlitSurface() failed: %s", SDL_GetError());
		}

		SDL_FreeSurface(lin);
	}

	SDL_Texture *tex = SDL_CreateTextureFromSurface(ren, sur);

	if (tex == NULL) {
		fail("SDL_CreateTextureFromSurface() failed: %s", SDL_GetError());
	}

	if (SDL_RenderCopy(ren, tex, NULL, NULL) < 0) {
		fail("SDL_RenderCopy() failed: %s", SDL_GetError());
	}

	for (int32_t y = 0; y < GFX_H * GFX_W; ++y) {
		for (int32_t p = 7; p > 1; --p) {
			uint32_t col = CON_BGR;
			if ((mem_gfx[y] & (1 << p)) > 0) {
				col = CON_DRW;
			}

			if (SDL_FillRect(gfx_sur, &(SDL_Rect){
				.x = y % 85 * fon_w + (8 - p) * fon_w / 6,
				.y = y / 85 * fon_h / 8,
				.w = fon_w / 6 + 1,
				.h = fon_h / 8 + 1
			}, col) < 0) {
				fail("SDL_FillRect() failed: %s", SDL_GetError());
			}
		}
	}

	SDL_Texture *gfx_tex = SDL_CreateTextureFromSurface(ren, gfx_sur);

	if (gfx_tex == NULL) {
		fail("SDL_CreateTextureFromSurface() failed: %s", SDL_GetError());
	}

	if (SDL_SetTextureBlendMode(gfx_tex, SDL_BLENDMODE_ADD) < 0) {
		fail("SDL_SetTextureBlendMode() failed: %s", SDL_GetError());
	}

	if (SDL_RenderCopy(ren, gfx_tex, NULL, NULL) < 0) {
		fail("SDL_RenderCopy() failed: %s", SDL_GetError());
	}

	SDL_DestroyTexture(tex);
	SDL_DestroyTexture(gfx_tex);
	SDL_RenderPresent(ren);
}

void lcd_init(void)
{
	ver("lcd init");

	win = SDL_CreateWindow("Front LCD", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED,
			WIN_W, WIN_H, 0);

	if (win == NULL) {
		fail("SDL_CreateWindow() failed: %s", SDL_GetError());
	}

	ren = SDL_CreateRenderer(win, -1, 0);

	if (ren == NULL) {
		fail("SDL_CreateRenderer() failed: %s", SDL_GetError());
	}

	SDL_AtomicSet(&clear, 0);

	SDL_AtomicSet(&frame, 1);

	SDL_RWops *ops = SDL_RWFromFile(font, "rb");

	if (ops == NULL) {
		fail("error while opening font file %s: %s", font, SDL_GetError());
	}

	fon = TTF_OpenFontRW(ops, 1, 32);

	if (fon == NULL) {
		fail("error while loading font file %s: %s", font, TTF_GetError());
	}

	fon_h = TTF_FontLineSkip(fon);

	if (TTF_GlyphMetrics(fon, 'X', NULL, NULL, NULL, NULL, &fon_w) < 0) {
		fail("error while measuring font width: %s", TTF_GetError());
	}

	sur_w = TXT_W * fon_w;
	sur_h = TXT_H * fon_h;

	sur = SDL_CreateRGBSurface(0, sur_w, sur_h, 32, 0, 0, 0, 0);

	if (sur == NULL) {
		fail("SDL_CreateRGBSurface() failed: %s", SDL_GetError());
	}

	gfx_sur = SDL_CreateRGBSurface(0, sur_w, sur_h, 32, 0, 0, 0, 0);

	if (gfx_sur == NULL) {
		fail("SDL_CreateRGBSurface() failed: %s", SDL_GetError());
	}

	for (int32_t y = 0; y < TXT_W * TXT_H; ++y) {
		mem_txt[y] = ' ';
	}
}

void lcd_quit(void)
{
	ver("lcd quit");

	SDL_FreeSurface(sur);
	SDL_FreeSurface(gfx_sur);
	TTF_CloseFont(fon);

	SDL_DestroyRenderer(ren);
	SDL_DestroyWindow(win);
}

bool lcd_exec(void)
{
	ver3("lcd exec");
	return false;
}

uint32_t lcd_read(uint32_t off, int32_t sz)
{
	ver2("lcd rd %u:%d", off, sz * 8);

	uint32_t rv;

	switch (current_op) {
		case G_MREAD:
			if (cur >= BASE_TXT && cur < BASE_TXT + TXT_W * TXT_H) {
				rv = mem_txt[cur - BASE_TXT];
			}
			else if (cur >= BASE_GFX && cur < BASE_GFX + GFX_W * GFX_H) {
				rv = mem_gfx[cur - BASE_GFX];
			}
			else {
				rv = 0x00;
			}
			break;

		default:
			rv = 0x00;
			break;
	}
	return rv;
}

void lcd_write(uint32_t off, int32_t sz, uint32_t val)
{
	if (last_val != val && mult_val_c > 0) {
		if (mult_val_c > 1) {
			ver2("lcd wr %u:%d 0x%0*x was called %u more times", last_off, last_sz * 8, last_sz * 2, last_val, mult_val_c);
		}
		else {
			ver2("lcd wr %u:%d 0x%0*x", last_off, last_sz * 8, last_sz * 2, last_val);
		}

		ver2("lcd wr %u:%d 0x%0*x", off, sz * 8, sz * 2, val);
		mult_val_c = 0;
	}
	else if (last_val == val && mult_val_c >= 0) {
		++mult_val_c;
	}
	else {
		ver2("lcd wr %u:%d 0x%0*x", off, sz * 8, sz * 2, val);
	}

	last_val = val;
	last_off = off;
	last_sz = sz;

	if (off == 0) {
		switch (current_op) {
			case G_MWRITE:
					if (cur >= BASE_TXT && cur < BASE_TXT + TXT_W * TXT_H) {
						if (val == 0) {
							mem_txt[cur - BASE_TXT] = ' ';
						}
						else {
							mem_txt[cur - BASE_TXT] = (uint8_t) val;
						}

						cur += dir;

						if (render) {
							SDL_AtomicAdd(&frame, 1);
						}
					}
					else if (cur >= BASE_GFX && cur < BASE_GFX + GFX_W * GFX_H) {
						mem_gfx[cur - BASE_GFX] = (uint8_t) val;
						cur += dir;

						if (render) {
							SDL_AtomicAdd(&frame, 1);
						}
					}
				break;

			case G_CRSWR:
				if (cur_c == 0) {
					cur = (int32_t) val;
					cur_c++;
				}
				else if (cur_c == 1) {
					cur = cur | ((int32_t) val << 8);
					if (cur < BASE_TXT || (cur >= BASE_TXT + TXT_W * TXT_H && cur < BASE_GFX) || cur >= BASE_GFX + GFX_W * GFX_H) {
						err("Invalid cur value %d", cur);
					}
					cur_c = 0;
				}
				break;

			default:
				break;
		}
	}
	else {
		switch (val) {
			case G_MWRITE:
				current_op = G_MWRITE;
				break;

			case G_CRSMRT:
				dir = 1;
				current_op = G_CRSMRT;
				break;

			case G_CRSMUP:
				dir = -85;
				current_op = G_CRSMUP;
				break;

			case G_CRSMDN:
				dir = 85;
				current_op = G_CRSMDN;
				break;

			case G_CRSWR:
				current_op = G_CRSWR;
				break;

			case G_MREAD:
				current_op = G_MREAD;
				break;

			case G_DSPOFF:
				SDL_AtomicSet(&clear, 1);
				render = 0;
				current_op = G_DSPOFF;
				break;

			case G_DSPON:
				render = 1;
				current_op = G_DSPON;
				break;

			default:
				current_op = 0x00;
				break;
		}
	}
}
